using MVR.FileManagementSecure;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;



public class SkinMicroDetail : MVRScript
{

    private DAZCharacterSelector selector;
    private DAZCharacter dazCharacter;


    private const string DEFAULT_DETAIL_MAP = "Custom/Atom/Person/Textures/Skin81.png";


    private static class BodyRegionEnum
    {
        public const string Torso = "Torso";
        public const string Face = "Face";
        public const string Limbs = "Limbs";
        public const string Genitals = "Genitals";
        public const string Nails = "Nails";
        public static readonly List<string> Values = new List<string>() { Torso, Genitals, Face, Limbs, Nails };
    }

    private static readonly Dictionary<string, int[]> femaleTextureIndex = new Dictionary<string, int[]>()
        {
            {BodyRegionEnum.Torso, new int[]{15,18,19,20,21,27,29}},
            {BodyRegionEnum.Face, new int[]{2,5,11} },
            {BodyRegionEnum.Limbs, new int[]{0,16,17,22,23}},
            {BodyRegionEnum.Genitals, new int[]{28}},
            {BodyRegionEnum.Nails, new int[]{12, 14}}
        };

    private static readonly Dictionary<string, int[]> maleTextureIndex = new Dictionary<string, int[]>()
        {
            {BodyRegionEnum.Torso, new int[]{15,16,19,20,21,22,30,32}},
            {BodyRegionEnum.Face, new int[]{2,5,11}},
            {BodyRegionEnum.Limbs, new int[]{0,17,18,23,24}},
            {BodyRegionEnum.Genitals, new int[]{28,29}},
            {BodyRegionEnum.Nails, new int[]{12, 14}}
    };


    private static readonly int[] maleSkinSlots =
        maleTextureIndex[BodyRegionEnum.Torso].Union(
        maleTextureIndex[BodyRegionEnum.Face]).Union(
        maleTextureIndex[BodyRegionEnum.Limbs]).Union(
        maleTextureIndex[BodyRegionEnum.Genitals]).Union(
        maleTextureIndex[BodyRegionEnum.Nails]).ToArray();

    private static readonly int[] femaleSkinSlots =
        femaleTextureIndex[BodyRegionEnum.Torso].Union(
        femaleTextureIndex[BodyRegionEnum.Face]).Union(
        femaleTextureIndex[BodyRegionEnum.Limbs]).Union(
        femaleTextureIndex[BodyRegionEnum.Genitals]).Union(
        femaleTextureIndex[BodyRegionEnum.Nails]).ToArray();

    private static readonly int[] nailSlots = new int[] { 12, 14 };

    private bool isLoaded = false;

    private Dictionary<int, Shader> originalShaders = new Dictionary<int, Shader>();
    // Some of the default skin materials do have tesselation already, so we need to be able to restore the original settings or we get seams
    private Dictionary<int, float> originalTesselation = new Dictionary<int, float>();
    private Dictionary<int, float> originalPhongStrength = new Dictionary<int, float>();
    private Dictionary<int, Texture> originalTessTexture = new Dictionary<int, Texture>();

    private Texture2D detailMap;

    private JSONStorableFloat tesselation = new JSONStorableFloat("Tesselation", 3.75f, 1.0f, 8.0f);
    private JSONStorableFloat phongStrength = new JSONStorableFloat("Phong Strength", 0.5f, 0.0f, 1.0f);
    private JSONStorableUrl detailMapURL = new JSONStorableUrl("Detail Map Texture", string.Empty);

    private JSONStorableFloat detailWeight = new JSONStorableFloat("Detail Weight", 0.25f, 0.0f, 1.0f);
    private JSONStorableFloat detailMapScaleTorso = new JSONStorableFloat("Detail Scale Torso", 50.0f, 0.0f, 200.0f, false);
    private JSONStorableFloat detailMapScaleFace = new JSONStorableFloat("Detail Scale Face", 50.0f, 0.0f, 200.0f, false);
    private JSONStorableFloat detailMapScaleLimbs = new JSONStorableFloat("Detail Scale Limbs", 50.0f, 0.0f, 200.0f, false);
    private JSONStorableFloat detailMapScaleGens = new JSONStorableFloat("Detail Scale Gens", 50.0f, 0.0f, 200.0f, false);

    // Distance Fading
    private JSONStorableBool enableDistanceFade = new JSONStorableBool("Enable Distance Fade", true);
    private JSONStorableStringChooser distanceFadeTargetControl;

    private JSONStorableFloat fadeMinDist = new JSONStorableFloat("Fade Min Distance", 1.0f, 0.0f, 100.0f, false);
    private JSONStorableFloat fadeMinDistDetail = new JSONStorableFloat("Fade Min Distance Detail Weight", 0.25f, 0.0f, 1.0f);

    private JSONStorableFloat fadeMaxDist = new JSONStorableFloat("Fade Max Distance", 1.75f, 0.0f, 100.0f, false);
    private JSONStorableFloat fadeMaxDistDetail = new JSONStorableFloat("Fade Max Distance Detail Weight", 0.0f, 0.0f, 1.0f);

    private JSONStorableBool preventShaderChange = new JSONStorableBool("Prevent Shader Overwrite (Compatibility with SSS Plug-in, etc.)", true);

    private Coroutine coroutine;
    private Transform distanceTarget;

    private Shader detailShader;
    private Shader tessShader;
    public override void Init()
    {


        if (containingAtom == null || containingAtom.type != "Person")
        {
            SuperController.LogError("This only works with person Atoms");
            DestroyImmediate(this);
            return;
        }



        selector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();
        dazCharacter = selector.selectedCharacter;

        RegisterFloat(tesselation);
        tesselation.setCallbackFunction += (float val) => UpdateShaderParams();
        CreateSlider(tesselation);
        RegisterFloat(phongStrength);
        phongStrength.setCallbackFunction += (float val) => UpdateShaderParams();
        CreateSlider(phongStrength);

        RegisterUrl(detailMapURL);
        detailMapURL.setCallbackFunction += LoadDetailMap;


        UIDynamicButton buttonSecondTex = CreateButton("Browse Detail Map", false);
        buttonSecondTex.button.onClick.AddListener(() =>
        {
            // Todo: Most recent DIR, or default
            List<ShortCut> shortCuts = FileManagerSecure.GetShortCutsForDirectory("Custom/Atom/Person/Textures", false, true, true, false);
            SuperController.singleton.GetMediaPathDialog((string path) => { detailMapURL.val = path; }, "jpg|png|tif|tiff", GetPackagePath(this) + "Custom/Atom/Person/Textures", false, true, false, null, true, shortCuts, true, true);

        });
        CreateTextField(detailMapURL, false);


        RegisterFloat(detailWeight);
        detailWeight.setCallbackFunction += (float val) => UpdateShaderParams();
        CreateSlider(detailWeight);

        RegisterFloat(detailMapScaleFace);
        detailMapScaleFace.setCallbackFunction += (float val) => UpdateShaderParams();
        CreateSlider(detailMapScaleFace);

        RegisterFloat(detailMapScaleTorso);
        detailMapScaleTorso.setCallbackFunction += (float val) => UpdateShaderParams();
        CreateSlider(detailMapScaleTorso);

        RegisterFloat(detailMapScaleLimbs);
        detailMapScaleLimbs.setCallbackFunction += (float val) => UpdateShaderParams();
        CreateSlider(detailMapScaleLimbs);

        RegisterFloat(detailMapScaleGens);
        detailMapScaleGens.setCallbackFunction += (float val) => UpdateShaderParams();
        CreateSlider(detailMapScaleGens);

        if (detailMapURL.val.Equals(string.Empty))
        {

            detailMapURL.SetFilePath(GetPackagePath(this) + DEFAULT_DETAIL_MAP);

        }

        RegisterBool(enableDistanceFade);

        CreateToggle(enableDistanceFade, true);



        List<string> controls = containingAtom.freeControllers.Select(control => control.name).ToList();
        distanceTarget = containingAtom.freeControllers.First(control => control.name == "hipControl").transform;

        distanceFadeTargetControl = new JSONStorableStringChooser("DistanceControlName", controls, "hipControl", "Distance Fade Target");
        distanceFadeTargetControl.setCallbackFunction += (val) =>
        {
            FreeControllerV3 controller = containingAtom.freeControllers.First(control => control.name == val);
            distanceTarget = controller ? controller.transform : null;
        };


        RegisterStringChooser(distanceFadeTargetControl);
        CreateFilterablePopup(distanceFadeTargetControl, true);





        RegisterFloat(fadeMaxDist);
        CreateSlider(fadeMaxDist, true);
        RegisterFloat(fadeMaxDistDetail);
        CreateSlider(fadeMaxDistDetail, true);

        RegisterFloat(fadeMinDist);
        CreateSlider(fadeMinDist, true);
        RegisterFloat(fadeMinDistDetail);
        CreateSlider(fadeMinDistDetail, true);



        RegisterBool(preventShaderChange);
        CreateToggle(preventShaderChange, true);
        detailShader = Shader.Find("Custom/Subsurface/GlossNMDetailTessMappedComputeBuff");
        tessShader = Shader.Find("Custom/Subsurface/GlossNMTessMappedFixedComputeBuff");
        coroutine = StartCoroutine(Setup());
    }

    private void LoadDetailMap(string path)
    {
        ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
        queuedImage.imgPath = path;
        queuedImage.forceReload = true;
        queuedImage.skipCache = true;
        queuedImage.isNormalMap = true;
        queuedImage.linear = true;
        queuedImage.compress = false;
        queuedImage.createMipMaps = false;
        queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
        {
            detailMap = qi.tex;
            UpdateShaderParams();
        };
        ImageLoaderThreaded.singleton.QueueImage(queuedImage);
    }

    private IEnumerator Setup()
    {
        while (!dazCharacter || !dazCharacter.ready || !detailShader || !tessShader)
        {
            yield return null; // Wait for a frame
        }

        foreach (int i in (GetMaterialSlots()))
        {
            Material m = dazCharacter.skin.GPUmaterials[i];
            if ((!nailSlots.Contains(i) && !detailShader.GetInstanceID().Equals(m.shader.GetInstanceID()))
                || (nailSlots.Contains(i) && !tessShader.GetInstanceID().Equals(m.shader.GetInstanceID())))
            {
 
                originalShaders[i] = m.shader;
                if (m.HasProperty("_Tess") && !originalTesselation.ContainsKey(i))
                {
                    originalTesselation[i] = m.GetFloat("_Tess");
                }
                if (m.HasProperty("_TessPhong") && !originalPhongStrength.ContainsKey(i))
                {
                    originalPhongStrength[i] = m.GetFloat("_TessPhong");
                }
                if (m.HasProperty("_TessTex") && !originalTessTexture.ContainsKey(i))
                {
                    originalTessTexture[i] = m.GetTexture("_TessTex");
                }
            }
        }

        foreach (int i in (GetMaterialSlots()))
        {
            Material mat = dazCharacter.skin.GPUmaterials[i];
            // Apply tesselation to nails, but with no skin detail
            if (nailSlots.Contains(i))
            {
                mat.shader = tessShader;
            } else
            {
                mat.shader = detailShader;
            }
        }

        UpdateShaderParams();
        dazCharacter.skin.BroadcastMessage("OnApplicationFocus", true);
        isLoaded = true;
    }

    public void Update()
    {

        if (enableDistanceFade.val && distanceTarget)
        {
            Vector3 cameraPosition = SuperController.singleton.centerCameraTarget.transform.position;
            float distance = Vector3.Distance(cameraPosition, distanceTarget.position);
            float t = Mathf.InverseLerp(fadeMinDist.val, fadeMaxDist.val, distance);
            detailWeight.val = Mathf.Lerp(fadeMinDistDetail.val, fadeMaxDistDetail.val, t);
        }

        if (selector && selector.selectedCharacter != dazCharacter)
        {
            dazCharacter = selector.selectedCharacter;
            if (coroutine != null)
            {
                StopCoroutine(coroutine);
            }

            originalShaders.Clear();
            originalPhongStrength.Clear();
            originalTesselation.Clear();
            originalTessTexture.Clear();

            coroutine = StartCoroutine(Setup());
        }

        // Detect Shader Reset on scene load, appearance preset load, etc.
        if (dazCharacter && dazCharacter.ready &&
            (!originalShaders.ContainsKey(0)
            || originalShaders[0].GetInstanceID() == dazCharacter.skin.GPUmaterials[0].shader.GetInstanceID())
            )
        {
            if (coroutine != null)
            {
                StopCoroutine(coroutine);
            }
            coroutine = StartCoroutine(Setup());
        }


        if (preventShaderChange.val && dazCharacter && dazCharacter.ready && detailShader.GetInstanceID() != dazCharacter.skin.GPUmaterials[0].shader.GetInstanceID())
        {
            if (coroutine != null)
            {
                StopCoroutine(coroutine);
            }
            coroutine = StartCoroutine(Setup());
        }


    }
    private void UpdateShaderParams()
    {
        if (!dazCharacter || !dazCharacter.ready)
        {
            return;
        }
        Dictionary<string, int[]> textureIndex = GetTextureIndex();
        foreach (int i in (GetMaterialSlots()))
        {
            Material mat = dazCharacter.skin.GPUmaterials[i];
            mat.SetTexture("_TessTex", Texture2D.whiteTexture);
            mat.SetFloat("_TessPhong", phongStrength.val);
            mat.SetFloat("_Tess", tesselation.val);
            mat.SetTexture("_DetailMap", detailMap);
            mat.SetFloat("_DetailWeight", detailWeight.val);
        }
        foreach (int i in textureIndex[BodyRegionEnum.Face])
        {
            Material mat = dazCharacter.skin.GPUmaterials[i];
            mat.SetTextureScale("_DetailMap", Vector2.one * detailMapScaleFace.val);
        }
        foreach (int i in textureIndex[BodyRegionEnum.Torso])
        {
            Material mat = dazCharacter.skin.GPUmaterials[i];
            mat.SetTextureScale("_DetailMap", Vector2.one * detailMapScaleTorso.val);
        }
        foreach (int i in textureIndex[BodyRegionEnum.Limbs])
        {
            Material mat = dazCharacter.skin.GPUmaterials[i];
            mat.SetTextureScale("_DetailMap", Vector2.one * detailMapScaleLimbs.val);
        }
        foreach (int i in textureIndex[BodyRegionEnum.Genitals])
        {
            Material mat = dazCharacter.skin.GPUmaterials[i];
            mat.SetTextureScale("_DetailMap", Vector2.one * detailMapScaleGens.val);
        }
    }

    private int[] GetMaterialSlots()
    {
        return dazCharacter.isMale ? maleSkinSlots : femaleSkinSlots;
    }

    private Dictionary<string, int[]> GetTextureIndex()
    {
        return dazCharacter.isMale ? maleTextureIndex : femaleTextureIndex;
    }

    public void OnDisable()
    {
        if (isLoaded && dazCharacter.ready)
        {
            foreach (int i in GetMaterialSlots())
            {
                dazCharacter.skin.GPUmaterials[i].shader = originalShaders[i];

                float value;
                if (originalTesselation.TryGetValue(i, out value))
                {
                    dazCharacter.skin.GPUmaterials[i].SetFloat("_Tess", value);
                }

                if (originalPhongStrength.TryGetValue(i, out value))
                {
                    dazCharacter.skin.GPUmaterials[i].SetFloat("_TessPhong", value);
                }
                Texture tex;
                if (originalTessTexture.TryGetValue(i, out tex))
                {
                    dazCharacter.skin.GPUmaterials[i].SetTexture("_TessTex", tex);
                }
            }
            dazCharacter.skin.BroadcastMessage("OnApplicationFocus", true);
        }
    }

    public void OnEnable()
    {
        if (coroutine != null)
        {
            StopCoroutine(coroutine);
        }
        coroutine = StartCoroutine(Setup());
    }

    public void OnDestroy()
    {
        OnDisable();
        StopAllCoroutines();
    }

    // Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
    // Credit to MacGruber Utils shared under CC BY
    private static string GetPluginPath(MVRScript self)
    {
        string id = self.name.Substring(0, self.name.IndexOf('_'));
        string filename = self.manager.GetJSON()["plugins"][id].Value;
        return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
    }

    // Get path prefix of the package that contains our plugin.
    // Credit to MacGruber Utils shared under CC BY
    private static string GetPackagePath(MVRScript self)
    {
        string filename = GetPluginPath(self);
        int idx = filename.IndexOf(":/");
        if (idx >= 0)
            return filename.Substring(0, idx + 2);
        else
            return string.Empty;
    }


}